@@ -565,6 +565,7 @@ PHP_APCU_API zend_bool apc_cache_store(apc_cache_t* cache, zend_string *strkey,
     return ret;
 } /* }}} */
 
+#ifndef ZTS
 /* {{{ data_unserialize */
 static zval data_unserialize(const char *filename)
 {
@@ -646,6 +647,7 @@ static int apc_load_data(apc_cache_t* cache, const char *data_file)
 
     return 0;
 }
+#endif
 
 /* {{{ apc_cache_preload shall load the prepared data files in path into the specified cache */
 PHP_APCU_API zend_bool apc_cache_preload(apc_cache_t* cache, const char *path)
@@ -1213,10 +1215,8 @@ static zval* my_serialize_object(zval* dst, const zval* src, apc_context_t* ctxt
 		}
 
 		ZVAL_STR(dst, serial);
-		/* Give this the type of an object/array, and the same type flags as a string. */
-		/* The only necessary one is probably IS_TYPE_REFCOUNTED - (We check Z_REFCOUNTED_P when de-duplicating serialized values. */
-		/* Probably safe - When copying strings into shared memory, the code gives it type IS_STRING_EX, */
-		Z_TYPE_INFO_P(dst) = Z_TYPE_P(src) | ((IS_TYPE_REFCOUNTED | IS_TYPE_COPYABLE) << Z_TYPE_FLAGS_SHIFT);
+		/* Give this the type of a refcounted object/array. */
+		Z_TYPE_INFO_P(dst) = Z_TYPE_P(src) | (IS_TYPE_REFCOUNTED << Z_TYPE_FLAGS_SHIFT);
 		efree(buf);
     }
 
@@ -1247,7 +1247,7 @@ static zval* my_unserialize_object(zval* dst, const zval* src, apc_context_t* ct
 
 static const uint32_t uninitialized_bucket[-HT_MIN_MASK] = {HT_INVALID_IDX, HT_INVALID_IDX};
 
-static zend_always_inline int apc_array_dup_element(apc_context_t *ctxt, HashTable *source, HashTable *target, uint32_t idx, Bucket *p, Bucket *q, int packed, int static_keys, int with_holes)
+static zend_always_inline int apc_array_dup_element(apc_context_t *ctxt, HashTable *source, HashTable *target, uint32_t idx, Bucket *p, Bucket *q, int packed, int with_holes)
 {
 	zval *data = &p->val;
 
@@ -1291,10 +1291,12 @@ static zend_always_inline int apc_array_dup_element(apc_context_t *ctxt, HashTab
 		uint32_t nIndex;
 
 		q->key = p->key;
-		if (!static_keys && q->key) {
+		if (q->key) {
 			if (ctxt->copy == APC_COPY_IN) {
 				q->key = apc_pstrcpy(q->key, ctxt->pool);
-			} else q->key = p->key;
+			} else {
+				q->key = zend_string_dup(p->key, 0);
+			}
 		}
 
 		nIndex = q->h | target->nTableMask;
@@ -1311,7 +1313,7 @@ static zend_always_inline void apc_array_dup_packed_elements(apc_context_t *ctxt
 	Bucket *end = p + source->nNumUsed;
 
 	do {
-		if (!apc_array_dup_element(ctxt, source, target, 0, p, q, 1, 1, with_holes)) {
+		if (!apc_array_dup_element(ctxt, source, target, 0, p, q, 1, with_holes)) {
 			if (with_holes) {
 				ZVAL_UNDEF(&q->val);
 			}
@@ -1320,7 +1322,7 @@ static zend_always_inline void apc_array_dup_packed_elements(apc_context_t *ctxt
 	} while (p != end);
 }
 
-static zend_always_inline uint32_t apc_array_dup_elements(apc_context_t *ctxt, HashTable *source, HashTable *target, int static_keys, int with_holes)
+static zend_always_inline uint32_t apc_array_dup_elements(apc_context_t *ctxt, HashTable *source, HashTable *target, int with_holes)
 {
     uint32_t idx = 0;
 	Bucket *p = source->arData;
@@ -1328,12 +1330,12 @@ static zend_always_inline uint32_t apc_array_dup_elements(apc_context_t *ctxt, H
 	Bucket *end = p + source->nNumUsed;
 
 	do {
-		if (!apc_array_dup_element(ctxt, source, target, idx, p, q, 0, static_keys, with_holes)) {
+		if (!apc_array_dup_element(ctxt, source, target, idx, p, q, 0, with_holes)) {
 			uint32_t target_idx = idx;
 
 			idx++; p++;
 			while (p != end) {
-				if (apc_array_dup_element(ctxt, source, target, target_idx, p, q, 0, static_keys, with_holes)) {
+				if (apc_array_dup_element(ctxt, source, target, target_idx, p, q, 0, with_holes)) {
 					if (source->nInternalPointer == idx) {
 						target->nInternalPointer = target_idx;
 					}
@@ -1361,7 +1363,11 @@ static APC_HOTSPOT HashTable* my_copy_hashtable(HashTable *source, apc_context_t
     if (target == NULL)
         goto bad;
 
+#if PHP_VERSION_ID >= 70300
+	GC_SET_REFCOUNT(target, 1);
+#else
 	GC_REFCOUNT(target) = 1;
+#endif
 	GC_TYPE_INFO(target) = IS_ARRAY;
     zend_hash_index_update_ptr(&ctxt->copied, (uintptr_t) source, target);
 
@@ -1369,39 +1375,23 @@ static APC_HOTSPOT HashTable* my_copy_hashtable(HashTable *source, apc_context_t
 	target->pDestructor = source->pDestructor;
 
 	if (source->nNumUsed == 0) {
+#if PHP_VERSION_ID >= 70300
+		target->u.flags = (source->u.flags & ~(HASH_FLAG_INITIALIZED|HASH_FLAG_PACKED)) | HASH_FLAG_STATIC_KEYS;
+#else
 		target->u.flags = (source->u.flags & ~(HASH_FLAG_INITIALIZED|HASH_FLAG_PACKED|HASH_FLAG_PERSISTENT)) | HASH_FLAG_APPLY_PROTECTION | HASH_FLAG_STATIC_KEYS;
+#endif
 		target->nTableMask = HT_MIN_MASK;
 		target->nNumUsed = 0;
 		target->nNumOfElements = 0;
 		target->nNextFreeElement = 0;
 		target->nInternalPointer = HT_INVALID_IDX;
 		HT_SET_DATA_ADDR(target, &uninitialized_bucket);
-	} else if (GC_FLAGS(source) & IS_ARRAY_IMMUTABLE) {
-		target->u.flags = (source->u.flags & ~HASH_FLAG_PERSISTENT) | HASH_FLAG_APPLY_PROTECTION;
-		target->nTableMask = source->nTableMask;
-		target->nNumUsed = source->nNumUsed;
-		target->nNumOfElements = source->nNumOfElements;
-		target->nNextFreeElement = source->nNextFreeElement;
-		if (ctxt->copy == APC_COPY_IN) {
-			HT_SET_DATA_ADDR(target, pool->palloc(pool, HT_SIZE(target)));
-		} else
-            HT_SET_DATA_ADDR(target, emalloc(HT_SIZE(target)));
-
-        if (HT_GET_DATA_ADDR(target) == NULL)
-            goto bad;
-
-		target->nInternalPointer = source->nInternalPointer;
-		memcpy(HT_GET_DATA_ADDR(target), HT_GET_DATA_ADDR(source), HT_USED_SIZE(source));
-		if (target->nNumOfElements > 0 &&
-			target->nInternalPointer == HT_INVALID_IDX) {
-			idx = 0;
-			while (Z_TYPE(target->arData[idx].val) == IS_UNDEF) {
-				idx++;
-			}
-			target->nInternalPointer = idx;
-		}
 	} else if (source->u.flags & HASH_FLAG_PACKED) {
+#if PHP_VERSION_ID >= 70300
+		target->u.flags = source->u.flags;
+#else
 		target->u.flags = (source->u.flags & ~HASH_FLAG_PERSISTENT) | HASH_FLAG_APPLY_PROTECTION;
+#endif
 		target->nTableMask = source->nTableMask;
 		target->nNumUsed = source->nNumUsed;
 		target->nNumOfElements = source->nNumOfElements;
@@ -1431,7 +1421,11 @@ static APC_HOTSPOT HashTable* my_copy_hashtable(HashTable *source, apc_context_t
 			target->nInternalPointer = idx;
 		}
 	} else {
-		target->u.flags = (source->u.flags & ~HASH_FLAG_PERSISTENT) | HASH_FLAG_APPLY_PROTECTION;
+#if PHP_VERSION_ID >= 70300
+		target->u.flags = (source->u.flags & ~HASH_FLAG_STATIC_KEYS);
+#else
+		target->u.flags = (source->u.flags & ~(HASH_FLAG_PERSISTENT|HASH_FLAG_STATIC_KEYS)) | HASH_FLAG_APPLY_PROTECTION;
+#endif
 		target->nTableMask = source->nTableMask;
 		target->nNextFreeElement = source->nNextFreeElement;
 		target->nInternalPointer = HT_INVALID_IDX;
@@ -1445,18 +1439,10 @@ static APC_HOTSPOT HashTable* my_copy_hashtable(HashTable *source, apc_context_t
 
 		HT_HASH_RESET(target);
 
-		if (target->u.flags & HASH_FLAG_STATIC_KEYS) {
-			if (source->nNumUsed == source->nNumOfElements) {
-				idx = apc_array_dup_elements(ctxt, source, target, 1, 0);
-			} else {
-				idx = apc_array_dup_elements(ctxt, source, target, 1, 1);
-			}
+		if (source->nNumUsed == source->nNumOfElements) {
+			idx = apc_array_dup_elements(ctxt, source, target, 0);
 		} else {
-			if (source->nNumUsed == source->nNumOfElements) {
-				idx = apc_array_dup_elements(ctxt, source, target, 0, 0);
-			} else {
-				idx = apc_array_dup_elements(ctxt, source, target, 0, 1);
-			}
+			idx = apc_array_dup_elements(ctxt, source, target, 1);
 		}
 		target->nNumUsed = idx;
 		target->nNumOfElements = idx;
@@ -1488,7 +1474,11 @@ static APC_HOTSPOT zend_reference* my_copy_reference(const zend_reference* src,
 	if (ctxt->copied.nTableSize) {
 		zend_reference *rc = zend_hash_index_find_ptr(&ctxt->copied, (uintptr_t) src);
 		if (rc) {
+#if PHP_VERSION_ID >= 70300
+			GC_ADDREF(rc);
+#else
 			GC_REFCOUNT(rc)++;
+#endif
 			return rc;
 		}
 	}
@@ -1502,11 +1492,16 @@ static APC_HOTSPOT zend_reference* my_copy_reference(const zend_reference* src,
     if (dst == NULL)
         return NULL;
 
+#if PHP_VERSION_ID >= 70300
+    GC_SET_REFCOUNT(dst, 1);
+#else
     GC_REFCOUNT(dst) = 1;
+#endif
     GC_TYPE_INFO(dst) = IS_REFERENCE;
 
-    if (my_copy_zval(&dst->val, &src->val, ctxt) == NULL)
+    if (my_copy_zval(&dst->val, &src->val, ctxt) == NULL) {
         return NULL;
+    }
 
 	if (ctxt->copied.nTableSize) {
 		zend_hash_index_update_ptr(&ctxt->copied, (uintptr_t) src, dst);
@@ -1560,10 +1555,12 @@ static APC_HOTSPOT zval* my_copy_zval(zval* dst, const zval* src, apc_context_t*
             return NULL;
 	    break;
 
+#ifdef IS_CONSTANT
     case IS_CONSTANT:
+#endif
     case IS_STRING:	
 		if (ctxt->copy == APC_COPY_OUT) {
-			ZVAL_DUP(dst, src);
+			ZVAL_STR(dst, zend_string_dup(Z_STR_P(src), 0));
 		} else {
 			Z_TYPE_INFO_P(dst) = IS_STRING_EX;
 			Z_STR_P(dst) = apc_pstrcpy(Z_STR_P(src), pool);
@@ -1574,8 +1571,11 @@ static APC_HOTSPOT zval* my_copy_zval(zval* dst, const zval* src, apc_context_t*
 
     case IS_ARRAY:
         if(ctxt->serializer == NULL) {
-			if ((Z_ARRVAL_P(dst) = my_copy_hashtable(Z_ARRVAL_P(src), ctxt)) == NULL)
-                return NULL;
+			HashTable *ht = my_copy_hashtable(Z_ARRVAL_P(src), ctxt);
+			if (!ht) {
+				return NULL;
+			}
+			ZVAL_ARR(dst, ht);
             break;
         }
 
@@ -1685,7 +1685,7 @@ static zval apc_cache_link_info(apc_cache_t *cache, apc_cache_slot_t* p)
 
     array_init(&link);
 
-    add_assoc_str(&link, "info", p->key.str);
+    add_assoc_str(&link, "info", zend_string_dup(p->key.str, 0));
     add_assoc_long(&link, "ttl", p->value->ttl);
 
     add_assoc_double(&link, "num_hits", (double)p->nhits);
@@ -1700,6 +1700,12 @@ static zval apc_cache_link_info(apc_cache_t *cache, apc_cache_slot_t* p)
 }
 /* }}} */
 
+#if APC_MMAP
+#define apc_cache_info_set_memory_type() add_assoc_stringl(&info, "memory_type", "mmap", sizeof("mmap")-1)
+#else
+#define apc_cache_info_set_memory_type() add_assoc_stringl(&info, "memory_type", "IPC shared", sizeof("IPC shared")-1)
+#endif
+
 /* {{{ apc_cache_info */
 PHP_APCU_API zval apc_cache_info(apc_cache_t* cache, zend_bool limited)
 {
@@ -1729,11 +1735,7 @@ PHP_APCU_API zval apc_cache_info(apc_cache_t* cache, zend_bool limited)
 		add_assoc_long(&info, "start_time", cache->header->stime);
 		add_assoc_double(&info, "mem_size", (double)cache->header->mem_size);
 
-#if APC_MMAP
-		add_assoc_stringl(&info, "memory_type", "mmap", sizeof("mmap")-1);
-#else
-		add_assoc_stringl(&info, "memory_type", "IPC shared", sizeof("IPC shared")-1);
-#endif
+		apc_cache_info_set_memory_type();
 
 		if (!limited) {
 		    /* For each hashtable slot */
@@ -1771,6 +1773,7 @@ PHP_APCU_API zval apc_cache_info(apc_cache_t* cache, zend_bool limited)
     return info;
 }
 /* }}} */
+#undef apc_cache_info_set_memory_type
 
 /*
  fetches information about the key provided
@@ -1873,8 +1876,24 @@ PHP_APCU_API void apc_cache_serializer(apc_cache_t* cache, const char* name) {
 	}
 } /* }}} */
 
+#ifndef APC_LOCK_RECURSIVE
+# define apc_cache_entry_try_begin() { \
+	if (APCG(recursion)++ == 0) { \
+		APC_LOCK(cache->header); \
+	} \
+}
 
-PHP_APCU_API void apc_cache_entry(apc_cache_t *cache, zval *key, zend_fcall_info *fci, zend_fcall_info_cache *fcc, zend_long ttl, zend_long now, zval *return_value) {
+# define apc_cache_entry_try_end() { \
+	if (--APCG(recursion) == 0) { \
+		APC_UNLOCK(cache->header); \
+	} \
+}
+#else
+# define apc_cache_entry_try_begin() APC_LOCK(cache->header);
+# define apc_cache_entry_try_end() APC_UNLOCK(cache->header);
+#endif
+
+PHP_APCU_API void apc_cache_entry(apc_cache_t *cache, zval *key, zend_fcall_info *fci, zend_fcall_info_cache *fcc, zend_long ttl, zend_long now, zval *return_value) {/*{{{*/
 	apc_cache_entry_t *entry = NULL;
 
 	if(!cache || apc_cache_busy(cache)) {
@@ -1886,15 +1905,7 @@ PHP_APCU_API void apc_cache_entry(apc_cache_t *cache, zval *key, zend_fcall_info
 		return;
 	}
 
-	php_apc_try({
-#ifndef APC_LOCK_RECURSIVE
-	if (APCG(recursion)++ == 0) {
-		APC_LOCK(cache->header);
-	}
-#else
-	APC_LOCK(cache->header);
-#endif
-	}, {
+	php_apc_try(apc_cache_entry_try_begin(), {
 		entry = apc_cache_find_internal(cache, Z_STR_P(key), now, 0);
 		if (!entry) {
 			int result = 0;
@@ -1915,16 +1926,10 @@ PHP_APCU_API void apc_cache_entry(apc_cache_t *cache, zval *key, zend_fcall_info
 				}
 			}
 		} else apc_cache_fetch_internal(cache, Z_STR_P(key), entry, now, &return_value);
-	}, {
-#ifndef APC_LOCK_RECURSIVE
-	if (--APCG(recursion) == 0) {
-		APC_UNLOCK(cache->header);
-	}
-#else
-	APC_UNLOCK(cache->header);
-#endif
-	});
-}
+	}, apc_cache_entry_try_begin());
+}/*}}}*/
+#undef apc_cache_entry_try_begin
+#undef apc_cache_entry_try_end
 
 /*
  * Local variables:
