@@ -43,7 +43,7 @@ static apc_iterator_item_t* apc_iterator_item_ctor(apc_iterator_t *iterator, apc
 
     array_init(&item->value);
 
-	item->key = slot->key.str;
+	item->key = zend_string_dup(slot->key.str, 0);
 
     if (APC_ITER_TYPE & iterator->format) {
 		add_assoc_string_ex(&item->value, "type", sizeof("type")-1, "user");
@@ -100,6 +100,7 @@ static zend_object* apc_iterator_clone(zval *zobject) {
 
 /* {{{ apc_iterator_item_dtor */
 static void apc_iterator_item_dtor(apc_iterator_item_t *item) {
+    zend_string_release(item->key);
     zval_ptr_dtor(&item->value);
     efree(item);
 }
@@ -123,6 +124,9 @@ static void apc_iterator_free(zend_object *object) {
 #ifdef ITERATOR_PCRE
     if (iterator->regex) {
         zend_string_release(iterator->regex);
+# if PHP_VERSION_ID >= 70300
+        pcre2_match_data_free(iterator->re_match_data);
+# endif
     }
 #endif
 
@@ -162,7 +166,11 @@ static int apc_iterator_search_match(apc_iterator_t *iterator, apc_cache_slot_t
 
 #ifdef ITERATOR_PCRE
     if (iterator->regex) {
+# if PHP_VERSION_ID >= 70300
+        rval = (pcre2_match(iterator->re, (PCRE2_SPTR) ZSTR_VAL((*slot)->key.str), ZSTR_LEN((*slot)->key.str), 0, 0, iterator->re_match_data, php_pcre_mctx()) >= 0);
+# else
         rval = (pcre_exec(iterator->re, NULL, ZSTR_VAL((*slot)->key.str), ZSTR_LEN((*slot)->key.str), 0, 0, NULL, 0) >= 0);
+# endif
     }
 #endif
 
@@ -197,7 +205,6 @@ static int apc_iterator_fetch_active(apc_iterator_t *iterator) {
     apc_cache_slot_t **slot;
     apc_iterator_item_t *item;
     time_t t;
-	zend_bool bailout = 0;
 
     t = apc_time();
 
@@ -205,9 +212,7 @@ static int apc_iterator_fetch_active(apc_iterator_t *iterator) {
         apc_iterator_item_dtor(apc_stack_pop(iterator->stack));
     }
 
-	APC_RLOCK(apc_user_cache->header);
-
-    zend_try {
+	php_apc_try(APC_RLOCK(apc_user_cache->header), {
 		while(count <= iterator->chunk_size && iterator->slot_idx < apc_user_cache->nslots) {
 		    slot = &apc_user_cache->slots[iterator->slot_idx];
 		    while(*slot) {
@@ -224,16 +229,10 @@ static int apc_iterator_fetch_active(apc_iterator_t *iterator) {
 		    }
 		    iterator->slot_idx++;
 		}
-	} zend_catch {
-		bailout = 1;
-	} zend_end_try();
-
-    iterator->stack_idx = 0;
-	APC_RUNLOCK(apc_user_cache->header);
-
-	if (bailout) {
-		zend_bailout();
-	}
+	}, {
+		iterator->stack_idx = 0; 
+		APC_RUNLOCK(apc_user_cache->header)
+	});
 
     return count;
 }
@@ -244,11 +243,8 @@ static int apc_iterator_fetch_deleted(apc_iterator_t *iterator) {
     int count=0;
     apc_cache_slot_t **slot;
     apc_iterator_item_t *item;
-	zend_bool bailout = 0;
 
-	APC_RLOCK(apc_user_cache->header);
-
-    zend_try {
+    php_apc_try(APC_RLOCK(apc_user_cache->header), {
 		slot = &apc_user_cache->header->gc;
 		while ((*slot) && count <= iterator->slot_idx) {
 		    count++;
@@ -265,17 +261,11 @@ static int apc_iterator_fetch_deleted(apc_iterator_t *iterator) {
 		    }
 		    slot = &(*slot)->next;
 		}
-	} zend_catch {
-		bailout = 1;
-	} zend_end_try();
-
-    iterator->slot_idx += count;
-    iterator->stack_idx = 0;
-	APC_RUNLOCK(apc_user_cache->header);
-
-	if (bailout) {
-		zend_bailout();
-	}
+	}, {
+		iterator->slot_idx += count;
+    	iterator->stack_idx = 0;
+		APC_RUNLOCK(apc_user_cache->header);
+	});
 
     return count;
 }
@@ -285,11 +275,8 @@ static int apc_iterator_fetch_deleted(apc_iterator_t *iterator) {
 static void apc_iterator_totals(apc_iterator_t *iterator) {
     apc_cache_slot_t **slot;
     int i;
-	zend_bool bailout = 0;
 
-	APC_RLOCK(apc_user_cache->header);
-
-    zend_try {
+    php_apc_try(APC_RLOCK(apc_user_cache->header), {
 		for (i=0; i < apc_user_cache->nslots; i++) {
 		    slot = &apc_user_cache->slots[i];
 		    while((*slot)) {
@@ -301,17 +288,10 @@ static void apc_iterator_totals(apc_iterator_t *iterator) {
 		        slot = &(*slot)->next;
 		    }
 		}
-	} zend_catch {
-		bailout = 1;
-	} zend_end_try();
-
-	APC_RUNLOCK(apc_user_cache->header);
-
-    iterator->totals_flag = 1;
-
-	if (bailout) {
-		zend_bailout();
-	}
+	}, {	
+		iterator->totals_flag = 1;
+		APC_RUNLOCK(apc_user_cache->header);
+	});
 }
 /* }}} */
 
@@ -357,15 +337,19 @@ void apc_iterator_obj_init(apc_iterator_t *iterator, zval *search, zend_long for
         iterator->regex = zend_string_copy(Z_STR_P(search));
         iterator->re = pcre_get_compiled_regex(iterator->regex, NULL, NULL);
 
-        if(!iterator->re) {
+        if (!iterator->re) {
             apc_error("Could not compile regular expression: %s", Z_STRVAL_P(search));
 			zend_string_release(iterator->regex);
+			iterator->regex = NULL;
         }
+
+# if PHP_VERSION_ID >= 70300
+        iterator->re_match_data = pcre2_match_data_create_from_pattern(iterator->re, php_pcre_gctx());
+# endif
 #else
         apc_error("Regular expressions support is not enabled, please enable PCRE for " APC_ITERATOR_NAME " regex support.");
 #endif
     } else if (search && Z_TYPE_P(search) == IS_ARRAY) {
-        Z_ADDREF_P(search);
         iterator->search_hash = apc_flip_hash(Z_ARRVAL_P(search));
     }
     iterator->initialized = 1;
